@@ -2,6 +2,7 @@ from typing import List, Optional, Dict, Any, Union
 
 from dataclasses import dataclass, field
 from code_loader.contract.enums import DatasetMetadataType, LeapDataType
+from code_loader.contract.mapping import NodeConnection
 
 
 @dataclass
@@ -20,6 +21,7 @@ class DatasetBaseSectionInstance:
 @dataclass
 class DatasetInputInstance(DatasetBaseSectionInstance):
     shape: List[int]
+    channel_dim: Optional[int] = -1
 
 
 @dataclass
@@ -95,7 +97,19 @@ class DatasetTestResultPayload:
 
 @dataclass
 class BoundingBox:
-    # (x, y) is the center of the bounding box
+    """
+    Represents a bounding box for an object in an image.
+
+    Attributes:
+    x (float): The x-coordinate of the center of the bounding box, a value between [0, 1] representing the percentage according to the image width.
+    y (float): The y-coordinate of the center of the bounding box, a value between [0, 1] representing the percentage according to the image height.
+    width (float): The width of the bounding box, a value between [0, 1] representing the percentage according to the image width.
+    height (float): The height of the bounding box, a value between [0, 1] representing the percentage according to the image height.
+    confidence (float): The confidence score of the bounding box. For predictions, this is a score typically between [0, 1]. For ground truth data, this can be 1.
+    label (str): The label or class name associated with the bounding box.
+    rotation (float): The rotation of the bounding box, a value between [0, 360] representing the degree of rotation. Default is 0.0.
+    metadata (Optional[Dict[str, Union[str, int, float]]]): Optional metadata associated with the bounding box.
+    """
     x: float  # value between [0, 1], represent the percentage according to the image size.
     y: float  # value between [0, 1], represent the percentage according to the image size.
 
@@ -107,6 +121,12 @@ class BoundingBox:
     metadata: Optional[Dict[str, Union[str, int, float]]] = None
 
 
+
+@dataclass
+class EngineFileContract:
+    node_connections: Optional[List[NodeConnection]] = None
+
+
 @dataclass
 class DatasetIntegParseResult:
     payloads: List[DatasetTestResultPayload]
@@ -116,3 +136,4 @@ class DatasetIntegParseResult:
     model_setup: Optional[ModelSetup] = None
     general_error: Optional[str] = None
     print_log: Optional[str] = None
+    engine_file_contract: Optional[EngineFileContract] = None
