@@ -1,5 +1,5 @@
 # mypy: ignore-errors
-
+import os
 from typing import Optional, Union, Callable, List, Dict
 
 import numpy as np
@@ -8,22 +8,129 @@ import numpy.typing as npt
 from code_loader.contract.datasetclasses import CustomCallableInterfaceMultiArgs, \
     CustomMultipleReturnCallableInterfaceMultiArgs, ConfusionMatrixCallableInterfaceMultiArgs, CustomCallableInterface, \
     VisualizerCallableInterface, MetadataSectionCallableInterface, PreprocessResponse, SectionCallableInterface, \
-    ConfusionMatrixElement, SamplePreprocessResponse, InstanceCallableInterface, ElementInstance
+    ConfusionMatrixElement, SamplePreprocessResponse, PredictionTypeHandler, InstanceCallableInterface, ElementInstance
 from code_loader.contract.enums import MetricDirection, LeapDataType, DatasetMetadataType
 from code_loader import leap_binder
 from code_loader.contract.mapping import NodeMapping, NodeMappingType, NodeConnection
 from code_loader.contract.visualizer_classes import LeapImage, LeapImageMask, LeapTextMask, LeapText, LeapGraph, \
     LeapHorizontalBar, LeapImageWithBBox, LeapImageWithHeatmap
+from code_loader.inner_leap_binder.leapbinder import mapping_runtime_mode_env_var_mame
+
+import inspect
+import functools
+
+_called_from_inside_tl_decorator = 0
+
+
+def _add_mapping_connection(user_unique_name, connection_destinations, arg_names, name, node_mapping_type):
+    connection_destinations = [connection_destination for connection_destination in connection_destinations
+                               if not isinstance(connection_destination, SamplePreprocessResponse)]
+
+    main_node_mapping = NodeMapping(name, node_mapping_type, user_unique_name, arg_names=arg_names)
+
+    node_inputs = {}
+    for arg_name, destination in zip(arg_names, connection_destinations):
+        node_inputs[arg_name] = destination.node_mapping
+
+    leap_binder.mapping_connections.append(NodeConnection(main_node_mapping, node_inputs))
 
 
 def _add_mapping_connections(connects_to, arg_names, node_mapping_type, name):
     for user_unique_name, connection_destinations in connects_to.items():
-        main_node_mapping = NodeMapping(name, node_mapping_type, user_unique_name, arg_names=arg_names)
-        node_inputs = {}
-        for arg_name, destination in zip(arg_names, connection_destinations):
-            node_inputs[arg_name] = destination.node_mapping
+        _add_mapping_connection(user_unique_name, connection_destinations, arg_names, name, node_mapping_type)
 
-        leap_binder.mapping_connections.append(NodeConnection(main_node_mapping, node_inputs))
+
+def integration_test():
+    def decorating_function(integration_test_function: Callable):
+        leap_binder.integration_test_func = integration_test_function
+
+        def inner(*args, **kwargs):
+            ret = integration_test_function(*args, **kwargs)
+
+            try:
+                os.environ[mapping_runtime_mode_env_var_mame] = 'True'
+                integration_test_function(None, None)
+            except Exception as e:
+                print(f'Error during integration test: Make sure to disable any non tensorleap decorators '
+                      f'functions before pushing a new TL version')
+            finally:
+                if mapping_runtime_mode_env_var_mame in os.environ:
+                    del os.environ[mapping_runtime_mode_env_var_mame]
+
+        return inner
+
+    return decorating_function
+
+
+def tensorleap_load_model(prediction_types: Optional[List[PredictionTypeHandler]] = None):
+    for i, prediction_type in enumerate(prediction_types):
+        leap_binder.add_prediction(prediction_type.name, prediction_type.labels, prediction_type.channel_dim, i)
+
+    def decorating_function(load_model_func):
+        class TempMapping:
+            pass
+
+        def inner():
+            class ModelPlaceholder:
+                def __init__(self):
+                    self.model = load_model_func()
+
+                # keras interface
+                def __call__(self, arg):
+                    ret = self.model(arg)
+                    return ret.numpy()
+
+                # onnx runtime interface
+                def run(self, output_names, input_dict):
+                    return self.model.run(output_names, input_dict)
+
+            return ModelPlaceholder()
+
+        def mapping_inner():
+            class ModelOutputPlaceholder:
+                def __init__(self):
+                    self.node_mapping = NodeMapping('', NodeMappingType.Prediction0)
+
+                def __getitem__(self, key):
+                    assert isinstance(key, int), \
+                        f'Expected key to be an int, got {type(key)} instead.'
+
+                    ret = TempMapping()
+                    ret.node_mapping = NodeMapping('', NodeMappingType(f'Prediction{str(key)}'))
+                    return ret
+
+            class ModelPlaceholder:
+                # keras interface
+                def __call__(self, arg):
+                    if isinstance(arg, list):
+                        for i, elem in enumerate(arg):
+                            elem.node_mapping.type = NodeMappingType[f'Input{str(i)}']
+                    else:
+                        arg.node_mapping.type = NodeMappingType.Input0
+
+                    return ModelOutputPlaceholder()
+
+                # onnx runtime interface
+                def run(self, output_names, input_dict):
+                    assert output_names is None
+                    assert isinstance(input_dict, dict), \
+                        f'Expected input_dict to be a dict, got {type(input_dict)} instead.'
+                    for i, elem in enumerate(input_dict.values()):
+                        elem.node_mapping.type = NodeMappingType[f'Input{str(i)}']
+
+                    return ModelOutputPlaceholder()
+
+            return ModelPlaceholder()
+
+        def final_inner():
+            if os.environ.get(mapping_runtime_mode_env_var_mame):
+                return mapping_inner()
+            else:
+                return inner()
+
+        return final_inner
+
+    return decorating_function
 
 
 def tensorleap_custom_metric(name: str,
@@ -38,12 +145,6 @@ def tensorleap_custom_metric(name: str,
                 raise Exception(f'Metric with name {name} already exists. '
                                 f'Please choose another')
 
-        leap_binder.add_custom_metric(user_function, name, direction, compute_insights)
-
-        if connects_to is not None:
-            arg_names = leap_binder.setup_container.metrics[-1].metric_handler_data.arg_names
-            _add_mapping_connections(connects_to, arg_names, NodeMappingType.Metric, name)
-
         def _validate_input_args(*args, **kwargs) -> None:
             for i, arg in enumerate(args):
                 assert isinstance(arg, (np.ndarray, SamplePreprocessResponse)), (
@@ -77,7 +178,8 @@ def tensorleap_custom_metric(name: str,
                             f'{supported_types_message}Got List[List[{type(single_metric_result[0][0])}]].'
                     else:
                         assert isinstance(single_metric_result[0], (
-                            float, int, type(None))), f'{supported_types_message}Got List[{type(single_metric_result[0])}].'
+                            float, int,
+                            type(None))), f'{supported_types_message}Got List[{type(single_metric_result[0])}].'
                 else:
                     assert isinstance(single_metric_result,
                                       np.ndarray), f'{supported_types_message}Got {type(single_metric_result)}.'
@@ -119,14 +221,59 @@ def tensorleap_custom_metric(name: str,
                         (f'tensorleap_custom_metric validation failed: '
                          f'compute_insights should be boolean. Got {type(compute_insights)}.')
 
+        @functools.wraps(user_function)
+        def inner_without_validate(*args, **kwargs):
+            global _called_from_inside_tl_decorator
+            _called_from_inside_tl_decorator += 1
+
+            try:
+                result = user_function(*args, **kwargs)
+            finally:
+                _called_from_inside_tl_decorator -= 1
+
+            return result
+
+        try:
+            inner_without_validate.__signature__ = inspect.signature(user_function)
+        except (TypeError, ValueError):
+            pass
+
+        leap_binder.add_custom_metric(inner_without_validate, name, direction, compute_insights)
+
+        if connects_to is not None:
+            arg_names = leap_binder.setup_container.metrics[-1].metric_handler_data.arg_names
+            _add_mapping_connections(connects_to, arg_names, NodeMappingType.Metric, name)
 
         def inner(*args, **kwargs):
             _validate_input_args(*args, **kwargs)
-            result = user_function(*args, **kwargs)
+
+            result = inner_without_validate(*args, **kwargs)
+
             _validate_result(result)
             return result
 
-        return inner
+        def mapping_inner(*args, **kwargs):
+            user_unique_name = mapping_inner.name
+            if 'user_unique_name' in kwargs:
+                user_unique_name = kwargs['user_unique_name']
+
+            ordered_connections = [kwargs[n] for n in mapping_inner.arg_names if n in kwargs]
+            ordered_connections = list(args) + ordered_connections
+            _add_mapping_connection(user_unique_name, ordered_connections, mapping_inner.arg_names,
+                                    mapping_inner.name, NodeMappingType.Metric)
+
+            return None
+
+        mapping_inner.arg_names = leap_binder.setup_container.metrics[-1].metric_handler_data.arg_names
+        mapping_inner.name = name
+
+        def final_inner(*args, **kwargs):
+            if os.environ.get(mapping_runtime_mode_env_var_mame):
+                return mapping_inner(*args, **kwargs)
+            else:
+                return inner(*args, **kwargs)
+
+        return final_inner
 
     return decorating_function
 
@@ -140,12 +287,6 @@ def tensorleap_custom_visualizer(name: str, visualizer_type: LeapDataType,
                 raise Exception(f'Visualizer with name {name} already exists. '
                                 f'Please choose another')
 
-        leap_binder.set_visualizer(user_function, name, visualizer_type, heatmap_function)
-
-        if connects_to is not None:
-            arg_names = leap_binder.setup_container.visualizers[-1].visualizer_handler_data.arg_names
-            _add_mapping_connections(connects_to, arg_names, NodeMappingType.Visualizer, name)
-
         def _validate_input_args(*args, **kwargs):
             for i, arg in enumerate(args):
                 assert isinstance(arg, (np.ndarray, SamplePreprocessResponse)), (
@@ -180,13 +321,59 @@ def tensorleap_custom_visualizer(name: str, visualizer_type: LeapDataType,
                 (f'tensorleap_custom_visualizer validation failed: '
                  f'The return type should be {result_type_map[visualizer_type]}. Got {type(result)}.')
 
+        @functools.wraps(user_function)
+        def inner_without_validate(*args, **kwargs):
+            global _called_from_inside_tl_decorator
+            _called_from_inside_tl_decorator += 1
+
+            try:
+                result = user_function(*args, **kwargs)
+            finally:
+                _called_from_inside_tl_decorator -= 1
+
+            return result
+
+        try:
+            inner_without_validate.__signature__ = inspect.signature(user_function)
+        except (TypeError, ValueError):
+            pass
+
+        leap_binder.set_visualizer(inner_without_validate, name, visualizer_type, heatmap_function)
+
+        if connects_to is not None:
+            arg_names = leap_binder.setup_container.visualizers[-1].visualizer_handler_data.arg_names
+            _add_mapping_connections(connects_to, arg_names, NodeMappingType.Visualizer, name)
+
         def inner(*args, **kwargs):
             _validate_input_args(*args, **kwargs)
-            result = user_function(*args, **kwargs)
+
+            result = inner_without_validate(*args, **kwargs)
+
             _validate_result(result)
             return result
 
-        return inner
+        def mapping_inner(*args, **kwargs):
+            user_unique_name = mapping_inner.name
+            if 'user_unique_name' in kwargs:
+                user_unique_name = kwargs['user_unique_name']
+
+            ordered_connections = [kwargs[n] for n in mapping_inner.arg_names if n in kwargs]
+            ordered_connections = list(args) + ordered_connections
+            _add_mapping_connection(user_unique_name, ordered_connections, mapping_inner.arg_names,
+                                    mapping_inner.name, NodeMappingType.Visualizer)
+
+            return None
+
+        mapping_inner.arg_names = leap_binder.setup_container.visualizers[-1].visualizer_handler_data.arg_names
+        mapping_inner.name = name
+
+        def final_inner(*args, **kwargs):
+            if os.environ.get(mapping_runtime_mode_env_var_mame):
+                return mapping_inner(*args, **kwargs)
+            else:
+                return inner(*args, **kwargs)
+
+        return final_inner
 
     return decorating_function
 
@@ -199,8 +386,6 @@ def tensorleap_metadata(
                 raise Exception(f'Metadata with name {name} already exists. '
                                 f'Please choose another')
 
-        leap_binder.set_metadata(user_function, name, metadata_type)
-
         def _validate_input_args(sample_id: Union[int, str], preprocess_response: PreprocessResponse):
             assert isinstance(sample_id, (int, str)), \
                 (f'tensorleap_metadata validation failed: '
@@ -228,9 +413,28 @@ def tensorleap_metadata(
                         (f'tensorleap_metadata validation failed: '
                          f'Values in the return dict should be of type {str(supported_result_types)}. Got {type(value)}.')
 
+        def inner_without_validate(sample_id, preprocess_response):
+
+            global _called_from_inside_tl_decorator
+            _called_from_inside_tl_decorator += 1
+
+            try:
+                result = user_function(sample_id, preprocess_response)
+            finally:
+                _called_from_inside_tl_decorator -= 1
+
+            return result
+
+        leap_binder.set_metadata(inner_without_validate, name, metadata_type)
+
         def inner(sample_id, preprocess_response):
+            if os.environ.get(mapping_runtime_mode_env_var_mame):
+                return None
+
             _validate_input_args(sample_id, preprocess_response)
-            result = user_function(sample_id, preprocess_response)
+
+            result = inner_without_validate(sample_id, preprocess_response)
+
             _validate_result(result)
             return result
 
@@ -261,7 +465,11 @@ def tensorleap_preprocess():
                  f'The return list should not contain duplicate PreprocessResponse objects.')
 
         def inner(*args, **kwargs):
+            if os.environ.get(mapping_runtime_mode_env_var_mame):
+                return [None, None, None, None]
+
             _validate_input_args(*args, **kwargs)
+
             result = user_function()
             _validate_result(result)
             return result
@@ -270,7 +478,9 @@ def tensorleap_preprocess():
 
     return decorating_function
 
-def tensorleap_element_instance_preprocess(instance_mask_encoder: Callable[[str, PreprocessResponse], List[ElementInstance]]):
+
+def tensorleap_element_instance_preprocess(
+        instance_mask_encoder: Callable[[str, PreprocessResponse], List[ElementInstance]]):
     def decorating_function(user_function: Callable[[], List[PreprocessResponse]]):
         def user_function_instance() -> List[PreprocessResponse]:
             result = user_function()
@@ -298,6 +508,7 @@ def tensorleap_element_instance_preprocess(instance_mask_encoder: Callable[[str,
 
         def builtin_instance_metadata(idx: str, preprocess: PreprocessResponse) -> Dict[str, str]:
             return {'is_instance': '0', 'original_sample_id': idx, 'instance_name': 'none'}
+
         leap_binder.set_preprocess(user_function_instance)
         leap_binder.set_metadata(builtin_instance_metadata, "builtin_instance_metadata")
 
@@ -319,7 +530,11 @@ def tensorleap_element_instance_preprocess(instance_mask_encoder: Callable[[str,
                  f'The return list should not contain duplicate PreprocessResponse objects.')
 
         def inner(*args, **kwargs):
+            if os.environ.get(mapping_runtime_mode_env_var_mame):
+                return [None, None, None, None]
+
             _validate_input_args(*args, **kwargs)
+
             result = user_function_instance()
             _validate_result(result)
             return result
@@ -356,8 +571,6 @@ def tensorleap_unlabeled_preprocess():
 
 def tensorleap_instances_masks_encoder(name: str):
     def decorating_function(user_function: InstanceCallableInterface):
-        leap_binder.set_instance_masks(user_function, name)
-
         def _validate_input_args(sample_id: str, preprocess_response: PreprocessResponse):
             assert isinstance(sample_id, str), \
                 (f'tensorleap_instances_masks_encoder validation failed: '
@@ -375,9 +588,27 @@ def tensorleap_instances_masks_encoder(name: str):
                 (f'tensorleap_instances_masks_encoder validation failed: '
                  f'Unsupported return type. Should be a numpy array. Got {type(result)}.')
 
+        def inner_without_validate(sample_id, preprocess_response):
+            global _called_from_inside_tl_decorator
+            _called_from_inside_tl_decorator += 1
+
+            try:
+                result = user_function(sample_id, preprocess_response)
+            finally:
+                _called_from_inside_tl_decorator -= 1
+
+            return result
+
+        leap_binder.set_instance_masks(inner_without_validate, name)
+
         def inner(sample_id, preprocess_response):
+            if os.environ.get(mapping_runtime_mode_env_var_mame):
+                return None
+
             _validate_input_args(sample_id, preprocess_response)
-            result = user_function(sample_id, preprocess_response)
+
+            result = inner_without_validate(sample_id, preprocess_response)
+
             _validate_result(result)
             return result
 
@@ -423,6 +654,9 @@ def tensorleap_input_encoder(name: str, channel_dim=-1, model_input_index=None):
             _validate_input_args(sample_id, preprocess_response)
             result = user_function(sample_id, preprocess_response)
             _validate_result(result)
+
+            if _called_from_inside_tl_decorator == 0:
+                result = np.expand_dims(result, axis=0)
             return result
 
         node_mapping_type = NodeMappingType.Input
@@ -430,7 +664,26 @@ def tensorleap_input_encoder(name: str, channel_dim=-1, model_input_index=None):
             node_mapping_type = NodeMappingType(f'Input{str(model_input_index)}')
         inner.node_mapping = NodeMapping(name, node_mapping_type)
 
-        return inner
+        def mapping_inner(sample_id, preprocess_response):
+            class TempMapping:
+                pass
+
+            ret = TempMapping()
+            ret.node_mapping = mapping_inner.node_mapping
+
+            return ret
+
+        mapping_inner.node_mapping = NodeMapping(name, node_mapping_type)
+
+        def final_inner(sample_id, preprocess_response):
+            if os.environ.get(mapping_runtime_mode_env_var_mame):
+                return mapping_inner(sample_id, preprocess_response)
+            else:
+                return inner(sample_id, preprocess_response)
+
+        final_inner.node_mapping = NodeMapping(name, node_mapping_type)
+
+        return final_inner
 
     return decorating_function
 
@@ -468,13 +721,33 @@ def tensorleap_gt_encoder(name: str):
             _validate_input_args(sample_id, preprocess_response)
             result = user_function(sample_id, preprocess_response)
             _validate_result(result)
+
+            if _called_from_inside_tl_decorator == 0:
+                result = np.expand_dims(result, axis=0)
             return result
 
         inner.node_mapping = NodeMapping(name, NodeMappingType.GroundTruth)
 
-        return inner
+        def mapping_inner(sample_id, preprocess_response):
+            class TempMapping:
+                pass
 
+            ret = TempMapping()
+            ret.node_mapping = mapping_inner.node_mapping
 
+            return ret
+
+        mapping_inner.node_mapping = NodeMapping(name, NodeMappingType.GroundTruth)
+
+        def final_inner(sample_id, preprocess_response):
+            if os.environ.get(mapping_runtime_mode_env_var_mame):
+                return mapping_inner(sample_id, preprocess_response)
+            else:
+                return inner(sample_id, preprocess_response)
+
+        final_inner.node_mapping = NodeMapping(name, NodeMappingType.GroundTruth)
+
+        return final_inner
 
     return decorating_function
 
@@ -486,13 +759,6 @@ def tensorleap_custom_loss(name: str, connects_to=None):
                 raise Exception(f'Custom loss with name {name} already exists. '
                                 f'Please choose another')
 
-        leap_binder.add_custom_loss(user_function, name)
-
-        if connects_to is not None:
-            arg_names = leap_binder.setup_container.custom_loss_handlers[-1].custom_loss_handler_data.arg_names
-            _add_mapping_connections(connects_to, arg_names, NodeMappingType.CustomLoss, name)
-
-
         valid_types = (np.ndarray, SamplePreprocessResponse)
         try:
             import tensorflow as tf
@@ -524,13 +790,63 @@ def tensorleap_custom_loss(name: str, connects_to=None):
                 (f'tensorleap_custom_loss validation failed: '
                  f'The return type should be a numpy array. Got {type(result)}.')
 
+        @functools.wraps(user_function)
+        def inner_without_validate(*args, **kwargs):
+            global _called_from_inside_tl_decorator
+            _called_from_inside_tl_decorator += 1
+
+            try:
+                result = user_function(*args, **kwargs)
+            finally:
+                _called_from_inside_tl_decorator -= 1
+
+            return result
+
+        try:
+            inner_without_validate.__signature__ = inspect.signature(user_function)
+        except (TypeError, ValueError):
+            pass
+
+        leap_binder.add_custom_loss(inner_without_validate, name)
+
+        if connects_to is not None:
+            arg_names = leap_binder.setup_container.custom_loss_handlers[-1].custom_loss_handler_data.arg_names
+            _add_mapping_connections(connects_to, arg_names, NodeMappingType.CustomLoss, name)
+
         def inner(*args, **kwargs):
             _validate_input_args(*args, **kwargs)
-            result = user_function(*args, **kwargs)
+
+            result = inner_without_validate(*args, **kwargs)
+
             _validate_result(result)
             return result
 
-        return inner
+        def mapping_inner(*args, **kwargs):
+            user_unique_name = mapping_inner.name
+            if 'user_unique_name' in kwargs:
+                user_unique_name = kwargs['user_unique_name']
+
+            ordered_connections = [kwargs[n] for n in mapping_inner.arg_names if n in kwargs]
+            ordered_connections = list(args) + ordered_connections
+            _add_mapping_connection(user_unique_name, ordered_connections, mapping_inner.arg_names,
+                                    mapping_inner.name, NodeMappingType.CustomLoss)
+
+            return None
+
+        mapping_inner.arg_names = leap_binder.setup_container.custom_loss_handlers[
+            -1].custom_loss_handler_data.arg_names
+        mapping_inner.name = name
+
+        def final_inner(*args, **kwargs):
+            if os.environ.get(mapping_runtime_mode_env_var_mame):
+                return mapping_inner(*args, **kwargs)
+            else:
+                return inner(*args, **kwargs)
+
+        final_inner.arg_names = leap_binder.setup_container.custom_loss_handlers[-1].custom_loss_handler_data.arg_names
+        final_inner.name = name
+
+        return final_inner
 
     return decorating_function
 
